-module(lazy2).
-compile(export_all).
-author('hanak@emt.bme.hu, kapolnai@iit.bme.hu').
-vsn('$LastChangedDate: 2020-11-17 07:14:09 +0100 (k, 17 nov 2020) $$').
-import('lazy',[cons/1,filter/2,map/2,take/2]).
-import('sendmory',[all_different/1,check_sum/1]).
  % A használatukhoz a lazy.beam és a sendmory.beam fájlnak léteznie kell
  % az adott mappában.

-type lazy_list() :: nil | {any(), fun(() -> lazy_list())}.
% Ez a lista félig lusta: a fej mindig kiértékelődik, a farok lusta.
% A teljesen lusta lista kicsit komplikáltabb, mert nem tudni, hogy
% üres vagy sem.

-spec next_variant(N::integer(), Ls::list()) -> R::{ok, Rs::list()} | nomore.
% Az Ls számsorozat után eggyel következő számsorozat Rs, ahol a számsorozat
% elemei 0 és N közé eső számok.
% Ha a számsorozat végére értünk, a függvény a nomore atomot adja eredményül.
% Előfeltétel: 0 <= Ls összes eleme <= N (a függvény nem ellenőrzi).
% Pl. N=2 mellett, ha Ls=[0,1], akkor Rs=[1,1]; ha Ls=[2,1], akkor Rs=[0,2].
next_variant(N, [N]) ->
    nomore;
next_variant(N, [N|T]) ->
    case next_variant(N, T) of
        {ok, C} -> {ok, [0|C]};
        nomore  -> nomore
    end;
next_variant(_N, [H|T]) ->
    {ok, [H+1|T]}.

-spec variants(N::integer(), Ls::list()) -> Ys::lazy_list().
% Az Ls számsorozattal kezdődő ismétléses variáció elemeinek lusta
% farkú listája Ys, ahol a számsorozatok elemei 0 és N közé eső számok.
% Előfeltétel: 0 <= Ls összes eleme <= N (a függvény nem ellenőrzi).
variants(N, Ls) ->
    {Ls, fun () -> case next_variant(N, Ls) of
		       {ok, Cs} -> variants(N, Cs);
		       nomore  -> nil
		   end
	 end
    }.

-spec pitag(N::integer()) -> Ys::lazy_list().
% Legfeljebb N összegű pitagoraszi számhármasok lusta farkú listája Ys.
pitag(N) ->
    filter(fun ([A,B,C]) ->
                   % io:format("~p~n",[{A,B,C}]),
                   A+B+C=<N andalso A=<B andalso A*B*C=/=0 andalso A*A+B*B=:=C*C
           end,
           variants(N,[2,2,3])).

-type d() :: 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9.
-type octet() :: {d(),d(),d(),d(),d(),d(),d(),d()}.
-spec smm(I::octet()) -> R::octet() | {}.
% A Send More Money feladvány I oktetet követő első megoldása az R oktet.
% Az I oktet csak a típusspecifikáció szerinti számjegyeket tartalmazhatja, de
% ezt a függvény nem ellenőrzi. 
% SEND+MORE=MONEY: 9567+1085=10652, mo: {S,E,N,D,M,O,R,Y} == {9,5,6,7,1,0,8,2}
smm(I) ->
    case filter(fun (L) ->
			%% io:format("~p~n",[L]),
			S = hd(L), M = lists:nth(5, L),
			S > 0 andalso M > 0 andalso
			    check_sum(list_to_tuple(L)) andalso all_different(L)
		end,
		variants(9,tuple_to_list(I)))
    of
	{Rs, _LazyTail} -> list_to_tuple(Rs)
      ;
	nil -> {}
    end.

test(0) ->
    [
     take(variants(3,[1,2]),2),
     take(variants(3,[1,2]),9),
     take(pitag(30), 2),
     take(pitag(30), 30),
     smm({0,0,0,0,0,0,2,2}) % Kis csalás, hogy gyorsabban megtalálja a megoldást.
    ];
test({_,_,_,_,_,_,_,_}=I) ->
    garbage_collect(), % prevent garbage collection during measurement,
    timer:sleep(500),  %   and wait for it a bit
    statistics(runtime), % clear runtime since last call
    Smm = smm(I),
    {_,T} = statistics(runtime), % get runtime since last call, ignore total runtime
    io:format("Result: ~w, time: ~5wms~n", [Smm, T]).
