% Hey Emacs! -*- coding: utf-8 -*-
-module(stack).
-compile(export_all).
-author('hanak@iit.bme.hu, modified by: kapolnai@iit.bme.hu').
-vsn('$LastChangedDate: 2019-10-29 21:47:55 +0100 (k, 29 okt 2019) $$').

-type stack() :: empty | {any(),stack()}.

-spec empty() -> S::stack().
% S az üres verem.
empty() -> empty.

-spec is_empty(S::stack()) -> B::boolean().
% B igaz, ha S üres.
is_empty(empty) -> true;
is_empty({_,_}) -> false.

-spec push(X::any(), S::stack()) -> S1::stack().
% Az S verem a tetején az X elemmel adja az S1 vermet.
push(X, empty)     -> {X,empty};
push(X, {_X,_S}=S) -> {X,S}.


-type stack_maybe() :: stack() | error.
-spec pop(S::stack()) -> S1::stack_maybe().
% Ha az S verem nem üres, a tetején lévő elem eltávolításával jön létre
% az S1 verem; ha üres, S1 =:= error.
pop(empty) -> error;
pop({_X,S}) -> S.

-type term_maybe() :: any() | error.
-spec top(S::stack()) -> X::term_maybe().
% Ha az S verem nem üres, a tetején lévő elem az X; ha üres, akkor X =:= error.
top(empty) -> error;
top({X,_S}) -> X.

-spec to_list(S::stack()) -> L::list().
% to_list(S) az S verem elemeit tartalmazó lista LIFO sorrendben.
to_list(empty) -> [];
to_list({X,S}) -> [X|to_list(S)].

test(1) ->
%    push(1,push(2,push(3,empty())));
    S1 = stack:push(1,stack:empty()),
    S2 = stack:push(2,S1),
    S3 = stack:push(3,S2),
    S3;
test(2) ->
    S = stack:push(3,stack:push(2,stack:push(1,stack:empty()))),
    io:format("stack: ~w, top: ~w, pop: ~w~n", [S, top(S),pop(S)]);
% szöveget megfordít.
test(3) ->
    Stack = lists:foldl(fun stack:push/2, stack:empty(), "szoveg"),
    to_list(Stack).

%% % @spec print(Stack) -> none().
%% % Kiírja a kimenetre a Stack tartalmát.
%% print(empty) ->
%%     io:format("empty~n");
%% print({X,S}) ->
%%     io:format("~p,", [X]),
%%     print(S).

